/*
 * pg_filedump.h - PostgreSQL file dump utility for dumping and
 *                 formatting heap(data), index and control files.
 *
 * Copyright (c) 2002, 2003 Red Hat, Inc. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Patrick Macdonald <patrickm@redhat.com> 
 *
 * Component of: PostgreSQL - Red Hat Edition Utilities / Tools
 *
 */

#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include "postgres.h"
#include "storage/bufpage.h"
#include "access/hash.h"
#include "access/gist.h"
#include "access/nbtree.h"
#include "access/rtree.h"
#include "access/itup.h"
#include "access/htup.h"
#include "catalog/pg_control.h"

// Options for Block formatting operations
static unsigned int blockOptions = 0;
typedef enum
{
  BLOCK_ABSOLUTE = 0x00000001,	// -a: Absolute (vs Relative) addressing
  BLOCK_BINARY = 0x00000002,	// -b: Binary dump of block
  BLOCK_FORMAT = 0x00000004,	// -f: Formatted dump of blocks / control file
  BLOCK_FORCED = 0x00000008,	// -S: Block size forced
  BLOCK_NO_INTR = 0x00000010,	// -d: Dump straight blocks
  BLOCK_RANGE = 0x00000020		// -R: Specific block range to dump
}
blockSwitches;

static int blockStart = -1;		// -R [start]: Block range start 
static int blockEnd = -1;		// -R [end]: Block range end

// Options for Item formatting operations
static unsigned int itemOptions = 0;
typedef enum
{
  ITEM_DETAIL = 0x00000001,		// -i: Display interpreted items    
  ITEM_HEAP = 0x00000002,		// -x: Blocks contain heap items
  ITEM_INDEX = 0x00000004		// -y: Blocks contain heap items
}
itemSwitches;

// Options for Control File formatting operations
static unsigned int controlOptions = 0;
typedef enum
{
  CONTROL_DUMP = 0x00000001,	// -c: Dump control file
  CONTROL_FORMAT = BLOCK_FORMAT,	// -f: Formatted dump of control file
  CONTROL_FORCED = BLOCK_FORCED	// -S: Block size forced
}
controlSwitches;

// Possible value types for the Special Section
static unsigned int specialType = 0;
typedef enum
{
  SPEC_SECT_NONE,				// No special section on block
  SPEC_SECT_SEQUENCE,			// Sequence info in special section 
  SPEC_SECT_INDEX,				// Index info in special section
  SPEC_SECT_ERROR_UNKNOWN,		// Unknown error 
  SPEC_SECT_ERROR_BOUNDARY		// Boundary error
}
specialSectionTypes;

// Possible return codes from option validation routine.
// pg_filedump doesn't do much with them now but maybe in
// the future...
typedef enum
{
  OPT_RC_VALID,					// All options are valid
  OPT_RC_INVALID,				// Improper option string
  OPT_RC_FILE,					// File problems
  OPT_RC_DUPLICATE,				// Duplicate option encountered
  OPT_RC_COPYRIGHT				// Copyright should be displayed
}
optionReturnCodes;

// Simple macro to check for duplicate options and then set
// an option flag for later consumption 
#define SET_OPTION(_x,_y,_z) if (_x & _y)               \
                               {                        \
                                 rc = OPT_RC_DUPLICATE; \
                                 duplicateSwitch = _z;  \
                               }                        \
                             else                       \
                               _x |= _y;

#define SEQUENCE_MAGIC 0x1717	// PostgreSQL defined magic number
#define EOF_ENCOUNTERED -1		// Indicator for partial read
#define BYTES_PER_LINE 16		// Format the binary 16 bytes per line


// --------------------------------------------------------------------
// THIS SECTION CONTAINS PRIOR VERSION DATA STRUCTURES ----------------
// --------------------------------------------------------------------

// This is the on-disk copy of the 7.1 / 7.2 heap tuple

typedef struct HeapTupleHeaderData_v0
{
  uint32 t_oid;					// OID of this tuple -- 4 bytes   
  uint32 t_cmin;				// insert CID stamp -- 4 bytes each   
  uint32 t_cmax;				// delete CommandId stamp   
  uint32 t_xmin;				// insert XID stamp -- 4 bytes each   
  uint32 t_xmax;				// delete XID stamp   
  ItemPointerData t_ctid;		// current TID of this or newer tuple   
  int16 t_natts;				// number of attributes   
  uint16 t_infomask;			// various infos   
  uint8 t_hoff;					// sizeof() tuple header   
  bits8 t_bits[1];				// bit map of NULLs   
}
HeapTupleHeaderData_v0;

typedef HeapTupleHeaderData_v0 *HeapTupleHeader_v0;

// The calculation for BITMAPLEN has change from PostgreSQL 7.2 to 7.3
#define MinHeapTupleBitmapSize_v0 32
#define BITMAPLEN_v0(NATTS) \
               ((((((int)(NATTS) - 1) >> 3) + 4 - (MinHeapTupleBitmapSize_v0 >> 3)) \
	         & ~03) + (MinHeapTupleBitmapSize_v0 >> 3))

// There are new fields in the control file.  This is the old one.      
typedef struct ControlFileData_v71
{
  crc64 crc;					// CRC for remainder of struct
  uint32 pg_control_version;	// PG_CONTROL_VERSION 
  uint32 catalog_version_no;	// see catversion.h
  DBState state;				// see enum above 
  time_t time;					// time stamp of last pg_control update 
  uint32 logId;					// current log file id 
  uint32 logSeg;				// current log file segment, + 1 
  XLogRecPtr checkPoint;		// last check point record ptr 
  XLogRecPtr prevCheckPoint;	// previous check point record ptr
  CheckPoint checkPointCopy;	// copy of last check point record
  uint32 blcksz;				// block size for this DB 
  uint32 relseg_size;			// blocks per segment of large relation 
  char lc_collate[LOCALE_NAME_BUFLEN];
  char lc_ctype[LOCALE_NAME_BUFLEN];
}
ControlFileData_v71;
